'use strict';

var _createClass = require('babel-runtime/helpers/create-class')['default'];

var _classCallCheck = require('babel-runtime/helpers/class-call-check')['default'];

var _extends = require('babel-runtime/helpers/extends')['default'];

var _slicedToArray = require('babel-runtime/helpers/sliced-to-array')['default'];

var _Object$entries = require('babel-runtime/core-js/object/entries')['default'];

var _regeneratorRuntime = require('babel-runtime/regenerator')['default'];

var _Promise = require('babel-runtime/core-js/promise')['default'];

var _Object$keys = require('babel-runtime/core-js/object/keys')['default'];

var _interopRequireDefault = require('babel-runtime/helpers/interop-require-default')['default'];

Object.defineProperty(exports, '__esModule', {
  value: true
});
exports.idFetcher = idFetcher;
exports.typeResolver = typeResolver;
exports.isConnection = isConnection;
exports.handleConnection = handleConnection;
exports.sequelizeNodeInterface = sequelizeNodeInterface;
exports.nodeAST = nodeAST;
exports.nodeType = nodeType;
exports.sequelizeConnection = sequelizeConnection;

var _graphqlRelay = require('graphql-relay');

var _graphql = require('graphql');

var _base64Js = require('./base64.js');

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _simplifyAST = require('./simplifyAST');

var _simplifyAST2 = _interopRequireDefault(_simplifyAST);

var NodeTypeMapper = (function () {
  function NodeTypeMapper() {
    _classCallCheck(this, NodeTypeMapper);

    this.map = {};
  }

  _createClass(NodeTypeMapper, [{
    key: 'mapTypes',
    value: function mapTypes(types) {
      var _this = this;

      _Object$entries(types).forEach(function (_ref) {
        var _ref2 = _slicedToArray(_ref, 2);

        var k = _ref2[0];
        var v = _ref2[1];

        _this.map[k] = v.type ? v : { type: v };
      });
    }
  }, {
    key: 'item',
    value: function item(type) {
      return this.map[type];
    }
  }]);

  return NodeTypeMapper;
})();

exports.NodeTypeMapper = NodeTypeMapper;

function idFetcher(sequelize, nodeTypeMapper) {
  var _this2 = this;

  return function callee$1$0(globalId, context) {
    var _fromGlobalId, type, id, nodeType, res, model;

    return _regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          _fromGlobalId = (0, _graphqlRelay.fromGlobalId)(globalId);
          type = _fromGlobalId.type;
          id = _fromGlobalId.id;
          nodeType = nodeTypeMapper.item(type);

          if (!(nodeType && typeof nodeType.resolve === 'function')) {
            context$2$0.next = 10;
            break;
          }

          context$2$0.next = 7;
          return _regeneratorRuntime.awrap(_Promise.resolve(nodeType.resolve(globalId, context)));

        case 7:
          res = context$2$0.sent;

          res.__graphqlType__ = type;
          return context$2$0.abrupt('return', res);

        case 10:
          model = _Object$keys(sequelize.models).find(function (model) {
            return model === type;
          });
          return context$2$0.abrupt('return', model ? sequelize.models[model].findById(id) : nodeType ? nodeType.type : null);

        case 12:
        case 'end':
          return context$2$0.stop();
      }
    }, null, _this2);
  };
}

function typeResolver(nodeTypeMapper) {
  return function (obj) {
    var type = obj.__graphqlType__ || (obj.Model ? obj.Model.options.name.singular : obj.name);

    if (!type) {
      throw new Error('Unable to determine type of ' + typeof obj + '. ' + 'Either specify a resolve function in \'NodeTypeMapper\' object, or specify \'__graphqlType__\' property on object.');
    }

    var nodeType = nodeTypeMapper.item(type);
    return nodeType && nodeType.type || null;
  };
}

function isConnection(type) {
  return typeof type.name !== 'undefined' && type.name.endsWith('Connection');
}

function handleConnection(values, args) {
  return (0, _graphqlRelay.connectionFromArray)(values, args);
}

function sequelizeNodeInterface(sequelize) {
  var nodeTypeMapper = new NodeTypeMapper();
  var nodeObjects = (0, _graphqlRelay.nodeDefinitions)(idFetcher(sequelize, nodeTypeMapper), typeResolver(nodeTypeMapper));

  return _extends({
    nodeTypeMapper: nodeTypeMapper
  }, nodeObjects);
}

function nodeAST(connectionAST) {
  return connectionAST.fields.edges && connectionAST.fields.edges.fields.node;
}

function nodeType(connectionType) {
  return connectionType._fields.edges.type.ofType._fields.node.type;
}

function sequelizeConnection(_ref3) {
  var name = _ref3.name;
  var nodeType = _ref3.nodeType;
  var target = _ref3.target;
  var orderByEnum = _ref3.orderBy;
  var _before = _ref3.before;
  var connectionFields = _ref3.connectionFields;
  var edgeFields = _ref3.edgeFields;
  var where = _ref3.where;

  var _connectionDefinitions = (0, _graphqlRelay.connectionDefinitions)({
    name: name,
    nodeType: nodeType,
    connectionFields: connectionFields,
    edgeFields: edgeFields
  });

  var edgeType = _connectionDefinitions.edgeType;
  var connectionType = _connectionDefinitions.connectionType;

  var model = target.target ? target.target : target;
  var SEPERATOR = '$';
  var PREFIX = 'arrayconnection' + SEPERATOR;

  if (orderByEnum === undefined) {
    orderByEnum = new _graphql.GraphQLEnumType({
      name: name + 'ConnectionOrder',
      values: {
        ID: { value: [model.primaryKeyAttribute, 'ASC'] }
      }
    });
  }

  _before = _before || function (options) {
    return options;
  };

  var $connectionArgs = _extends({}, _graphqlRelay.connectionArgs, {
    orderBy: {
      type: new _graphql.GraphQLList(orderByEnum)
    }
  });

  var orderByAttribute = function orderByAttribute(orderBy) {
    return orderBy[0][0];
  };

  /**
   * Creates a cursor given a item returned from the Database
   * @param  {Object}   item   sequelize model instance
   * @param  {Integer}  index  the index of this item within the results, 0 indexed
   * @return {String}          The Base64 encoded cursor string
   */
  var toCursor = function toCursor(item, index) {
    var id = item.get(model.primaryKeyAttribute);
    return (0, _base64Js.base64)(PREFIX + id + SEPERATOR + index);
  };

  /**
   * Decode a cursor into its component parts
   * @param  {String} cursor Base64 encoded cursor
   * @return {Object}        Object containing ID and index
   */
  var fromCursor = function fromCursor(cursor) {
    cursor = (0, _base64Js.unbase64)(cursor);
    cursor = cursor.substring(PREFIX.length, cursor.length);

    var _cursor$split = cursor.split(SEPERATOR);

    var _cursor$split2 = _slicedToArray(_cursor$split, 2);

    var id = _cursor$split2[0];
    var index = _cursor$split2[1];

    return {
      id: id,
      index: index
    };
  };

  var argsToWhere = function argsToWhere(args) {
    var result = {};

    _lodash2['default'].each(args, function (value, key) {
      if (key in $connectionArgs) return;
      _lodash2['default'].assign(result, where(key, value));
    });

    return result;
  };

  var resolveEdge = function resolveEdge(item, index, queriedCursor, args, source) {
    if (args === undefined) args = {};

    var startIndex = 0;
    if (queriedCursor) startIndex = Number(queriedCursor.index);
    if (startIndex !== 0) startIndex++;
    return {
      cursor: toCursor(item, index + startIndex),
      node: item,
      source: source
    };
  };

  var $resolver = require('./resolver')(target, {
    handleConnection: false,
    include: true,
    list: true,
    before: function before(options, args, context) {
      if (args.first || args.last) {
        options.limit = parseInt(args.first || args.last, 10);
      }
      if (!args.orderBy) {
        args.orderBy = [orderByEnum._values[0].value];
      } else if (typeof args.orderBy === 'string') {
        args.orderBy = [orderByEnum._nameLookup[args.orderBy].value];
      }

      var orderBy = args.orderBy;
      var orderAttribute = orderByAttribute(orderBy);
      var orderDirection = args.orderBy[0][1];

      if (args.last) {
        orderDirection = orderDirection === 'ASC' ? 'DESC' : 'ASC';
      }

      options.order = [[orderAttribute, orderDirection]];

      if (orderAttribute !== model.primaryKeyAttribute) {
        options.order.push([model.primaryKeyAttribute, 'ASC']);
      }

      options.attributes.push(orderAttribute);

      if (model.sequelize.dialect.name === 'postgres' && options.limit) {
        options.attributes.push([model.sequelize.literal('COUNT(*) OVER()'), 'full_count']);
      } else if (model.sequelize.dialect.name === 'mssql' && options.limit) {
        options.attributes.push([model.sequelize.literal('COUNT(1) OVER()'), 'full_count']);
      }

      options.where = argsToWhere(args);
      options.required = false;

      if (args.after || args.before) {
        var cursor = fromCursor(args.after || args.before);
        var startIndex = Number(cursor.index);

        if (startIndex > 0) options.offset = startIndex + 1;
      }
      options.attributes = _lodash2['default'].uniq(options.attributes);
      return _before(options, args, root, context);
    },
    after: function after(values, args, root, _ref4) {
      var source = _ref4.source;

      var cursor = null;

      if (args.after || args.before) {
        cursor = fromCursor(args.after || args.before);
      }

      var edges = values.map(function (value, idx) {
        return resolveEdge(value, idx, cursor, args, source);
      });

      var firstEdge = edges[0];
      var lastEdge = edges[edges.length - 1];
      var fullCount = values[0] && values[0].dataValues.full_count && parseInt(values[0].dataValues.full_count, 10);

      if (!values[0]) {
        fullCount = 0;
      }
      if (model.sequelize.dialect.name === 'postgres' && (args.first || args.last)) {
        if (fullCount === null || fullCount === undefined) throw new Error('No fullcount available');
      }
      var hasMorePages = false;
      if (args.first || args.last) {
        var index = cursor ? Number(cursor.index) : 0;
        hasMorePages = index + 1 + parseInt(args.first || args.last, 10) < fullCount;
      }

      return {
        source: source,
        args: args,
        where: argsToWhere(args),
        edges: edges,
        pageInfo: {
          startCursor: firstEdge ? firstEdge.cursor : null,
          endCursor: lastEdge ? lastEdge.cursor : null,
          hasPreviousPage: hasMorePages,
          hasNextPage: hasMorePages
        }
      };
    }
  });

  var resolver = function resolver(source, args, context, info) {
    if ((0, _simplifyAST2['default'])(info.fieldASTs[0], info).fields.edges) {
      return $resolver(source, args, context, info);
    }

    return {
      source: source,
      args: args,
      where: argsToWhere(args)
    };
  };

  resolver.$association = $resolver.$association;
  resolver.$before = $resolver.$before;
  resolver.$after = $resolver.$after;
  resolver.$options = $resolver.$options;

  return {
    connectionType: connectionType,
    edgeType: edgeType,
    nodeType: nodeType,
    resolveEdge: resolveEdge,
    connectionArgs: $connectionArgs,
    resolve: resolver
  };
}